/*' $Header:   P:/PVCS/MAX/QUILIB/TRUENAME.C_V   1.3   30 May 1997 12:09:06   BOB  $ */
/******************************************************************************
 *									      *
 * (C) Copyright 1991-92 Qualitas, Inc.  All rights reserved.		      *
 *									      *
 * TRUENAME.C								      *
 *									      *
 * Functions for canonical filename/path processing			      *
 *									      *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/******************** TRUENAME_FN (char far *path) ***************************/
// Returns dst_path if TRUENAME call succeeded or NULL if it failed

#pragma optimize ("leg",off)    /* Global optimization precludes ... */
char far *truename_fn (char far *dst_path, char far *src_path)
{
 _asm {

	push	si		// Save SI
	push	di		// Save DI
	push	ds		// Save DS
	push	es		// Save ES

	lds	si,src_path	// Path to convert
	les	di,dst_path	// Destination for converted path

	// WARNING: This call does not get properly handled under certain
	// as yet not fully understood circumstances in Windows...
	mov	ah,0x60 	// DOS function 60h (TRUENAME)
	int	0x21		// Call DOS

	pop	es		// Restore
	pop	ds
	pop	di
	pop	si

	jnc	Truename_OK	// No errors

 }
	return (NULL);		// Indicate failure

Truename_OK:
	return (dst_path);	// Indicate success

} /* truename_fn () */
#pragma optimize ("",on)

/********************* TRUENAME (char *path) *****************************/
// Convert a fully qualified pathname to its actual name via DOS truename
// function (AH = 60h, ds:si ==> source, es:di ==> dest; maximum length
// on output = 128).  *** NOTE *** The names generated by this function may
// not be valid with the current set of SUBST's and JOIN's.
void truename (char *path)
{
 char temp[_MAX_PATH];

 if (truename_fn (temp, path)) {	// Call succeeded
	strcpy (path, temp);		// Copy canonical form over original
 }

} // truename ()

/****************** TRUENAME_CMP () **********************************/
/* Returns strcmp() results for path1 and path2 after they have been */
/* laundered through truename() */

int truename_cmp (char *path1, char *path2)
{
 char	temp1[_MAX_PATH],
	temp2[_MAX_PATH];

 strcpy (temp1, path1);
 strcpy (temp2, path2);

 truename (temp1);
 truename (temp2);

 return (_stricmp (temp1, temp2));

} /* truename_cmp () */

/******************* TRUEDIR () ***************************************/
/* Takes a backslash-terminated [drive:]directory and converts it to */
/* canonical form. */

void truedir (char *dir)
{
 char	temp1[_MAX_PATH],
	temp2[_MAX_PATH];
 int	i;

 sprintf (temp1, "%s?.?", dir);
 if (truename_fn (temp2, temp1)) {
	i = strlen (temp2) - 3; 	// Omit "?.?"
	strncpy  (dir, temp2, i);
	dir [i] = '\0';
 }

} /* truedir () */

